import gobject
from units import Soldier, Archer, Saboteur
import globals

class Board(gobject.GObject):
    "Board of the game, which contain all the units."
    __gsignals__ = {
        "unit_added" : (gobject.SIGNAL_RUN_FIRST, None, [int,int,object]),
        "unit_removed" : (gobject.SIGNAL_RUN_FIRST, None, [int,int])
        }

    def __init__ (self, w, h):
        gobject.GObject.__init__(self) # Call the parent constructeur
        self._w = w
        self._h = h
        self.board = None
        self.create()

    def create(self):
        "creation of the board"

        self.board = [globals.EMPTY] * self._w
        for  w in range(0, self._w): # we fill all the column of the board
            self.board[w] = [globals.EMPTY] * self._h

    def clean(self):
        "clean the board"
        for x in range(self._w):
            for y in range(self._h):
                if not self.empty(x, y):
                    self.remove_unit(x, y)

    def empty(self, x, y):
        "return if a case is empty"
        return (self.board[x][y] == globals.EMPTY)

    def add_unit(self, x, y, unit):
        "add a unit in a certain place of the board"
        if self.board[x][y] != globals.EMPTY:
            raise globals.CaseNotEmpty(x, y)
        else:
            self.board[x][y] = unit
            self.emit("unit_added", x, y, unit)

    def remove_unit (self, x, y):
        "remove a unit in a certain place of the board"
        if self.board[x][y] == globals.EMPTY:
            raise globals.CaseIsEmpty(x, y)
        else:
            toreturn = self.board[x][y]
            self.board[x][y] = globals.EMPTY

            self.emit("unit_removed", x, y)
            return toreturn

    def is_valid(self, x, y):
        return x >= 0 and y >= 0 and y < self._w and y < self._h

class Game(gobject.GObject):
    "main class of the game"
    # Create some signal, which are fired when the state of the game change
    __gsignals__ = {
        # team which won
        "player_wins" : (gobject.SIGNAL_RUN_FIRST, None, [str]),
        #  New game started
        "new_game" : (gobject.SIGNAL_RUN_FIRST, None, []),
        # playing team, turn number
        "turn_ended" : (gobject.SIGNAL_RUN_FIRST, None, [str, int]),
        # unit, team, x, y
        "initial_placement": (gobject.SIGNAL_RUN_FIRST, None, [object, int,
            int]),
        # unit, to_x, to_y, from_x, from_y
        "unit_moved" : (gobject.SIGNAL_RUN_LAST, None, [object, int, int, int,
            int]),
        # attacker, to_x, to_y, victim, from_x, from_y
        "direct_attack" : (gobject.SIGNAL_RUN_LAST, None, [object, int, int,
            object, int, int]),
        # attacker, to_x, to_y, victim, from_x, from_y
        "ranged_attack" : (gobject.SIGNAL_RUN_LAST, None, [object, int, int,
            object, int, int]),
        # unit, to_x, to_y, from_x, from_y
        "saboteur_return": (gobject.SIGNAL_RUN_LAST, None, [object, int, int,
            int, int])
        }

    initial_positions = {
            "Black" : {
                Soldier: [(22, 1), (22, 2), (22, 3), (22, 4)],
                Archer: [(25,2), (25,3)],
                Saboteur: [(26,1), (26,2), (26,3), (26,4)]
            },
            "White": {
                Soldier: [(4,1), (4,2), (4,3), (4,4)],
                Archer: [(1,2), (1,3)],
                Saboteur: [(0,1), (0,2), (0,3), (0,4)]
            }
    }

    def __init__(self):
        gobject.GObject.__init__(self)
        self.board = Board(globals.TOTAL_WIDTH, globals.TOTAL_HEIGHT)

    def get_unit(self, x, y):
        return self.board.board[x][y]

    def new_game(self):
        "create a new game"
        self.emit("new_game")
        self.board.clean()
        self._initial_placement("White")
        self._initial_placement("Black")

        self.invasion_positions = {
            "White" : [(24, 1), (24, 2), (24,3), (24,4)],
            "Black" : [(2, 1), (2, 2), (2,3), (2,4)]
        }

        self.has_moved = []
        self._has_invaded = []
        # Should this be random?
        self.turn = "White"
        self.who_start = "White"
        self.turn_number = globals.FIRST_TURN
        self.chosen_type_of_units = []

        self.emit("turn_ended", self.turn, self.turn_number)

    def end_turn (self):
        "give hand to the other player"

        self.has_moved = []
        self.chosen_type_of_units = []

        # toggle sides
        if self.turn == "Black":
            self.turn = "White"
        else:
            self.turn = "Black"

        # inc turn number
        if self.turn == self.who_start:
            self.turn_number += globals.TURN

        # Notify the interface than the turn ended
        self.emit("turn_ended", self.turn, self.turn_number)

    def can_attack_from (self, x, y):
        "Return the list of position where a unit can attack"
        unit = self.get_unit(x, y)

        # if we are in the battlefield
        if globals.is_in_battlefield(x, y):
            positions = unit.can_attack_from(x, y, self.board)
        else:
            # Don't allow attack from the bases
            #See Bug #337660 https://bugs.launchpad.net/pyhalebarde/+bug/337660
            positions = []

        return positions

    def can_go_from (self, from_x, from_y):
        "return the list of position where a unit can move"
        unit = self.get_unit(from_x, from_y)
        positions = unit.can_go_from(from_x, from_y, self.board)
        return positions

    def _return_to_base(self, saboteur):
        "if a saboteur is killed he has to return to its base"
        to_x, to_y = 0, 0
        # for all the 4 available place in the base
        for x, y in self.initial_positions[saboteur.side][saboteur.__class__]:
            if self.board.empty(x, y):
                to_x, to_y = x, y

        if (to_x, to_y) != (0, 0):
            self.board.add_unit(to_x, to_y, saboteur)
        else:
            raise globals.BaseIsAnormalyFull

        return to_x, to_y

    def _enter_enemy_base(self, saboteur):
        "place a saboteur unit in the enemy base"
        x, y = self.invasion_positions[saboteur.side].pop()
        self.board.add_unit(x, y, saboteur)
        print "Saboteur invaded at (%i,%i)" % (x, y)
        self._has_invaded.append(saboteur)

        # If all positions are occupied, the side has win
        has_win = False
        if self.invasion_positions[saboteur.side] == []:
            has_win = saboteur.side
        return has_win, x, y

    def _initial_placement(self, team):
        "Insert the beginning units for a given team in its base"
        units = self.initial_positions[team]
        for unit_class, positions in units.items():
            for x, y in positions:
                unit = unit_class(team)
                self.board.add_unit(x, y, unit)
                self.emit("initial_placement", unit, x, y)

    def can_act(self, x, y):
        "Check if a unit can act (do a movement or attack) during this turn"
        unit = self.get_unit(x, y)
        if not unit:
            return False

        # Fix #336355 : Don't allow saboteur to move in enemy base.
        if unit in self._has_invaded:
            return False

        # Follow the complicated rules to see if an unit can move.
        return self.turn == unit.side\
                and unit not in self.has_moved\
                and (self.turn_number == globals.FIRST_TURN\
                    or unit.__class__ in self.chosen_type_of_units\
                    or len(self.chosen_type_of_units) <\
                        globals.MAX_PLAYABLE_UNIT_TYPE)

    def _has_acted(self, unit):
        "Mark an unit as having moved or attacked this turn"
        self.has_moved.append(unit)
        if unit.__class__ not in self.chosen_type_of_units:
            self.chosen_type_of_units.append(unit.__class__)

    def attack(self, x, y, from_x, from_y):
        "attack a unit with an other unit"
        assailant = self.get_unit(from_x,from_y)
        victim = self.board.remove_unit(x, y)

        # If attacking unit is a Soldier, it's a direct attack
        # and it take the place of the attacked unit
        if assailant.__class__ == Soldier:
            self.board.remove_unit(from_x, from_y)
            self.board.add_unit(x, y, assailant)
            self.emit("direct_attack", assailant, x, y,
                    victim, from_x, from_y)
        else: #It's an archer who is attacking then ranged attack
            self.emit("ranged_attack", assailant, x, y, victim,
                    from_x, from_y)

        # If attacked unit is a Saboteur it return to his base, otherwise it
        # just die
        if victim.__class__ == Saboteur:
            rx, ry = self._return_to_base(victim)
            self.emit("saboteur_return", victim, rx, ry, x, y)

        self._has_acted(assailant)

    def move(self, x, y, from_x, from_y):
        "move a unit from a position to an other"

        unit = self.get_unit(from_x, from_y)
        has_win = None
        # If the unit is a saboteur, see if it has reached the enemy base.
        if unit.__class__ == Saboteur and \
            ((unit.side == "Black" and (x, y) in globals.WHITE_BASE_ENTRY) or\
            (unit.side == "White" and (x, y) in globals.BLACK_BASE_ENTRY)):
            # if a saboteur has reached the enemy base
            has_win, x, y = self._enter_enemy_base(unit)
        else:
            self.board.add_unit(x, y, unit)

        self.board.remove_unit(from_x, from_y)

        self._has_acted(unit)

        self.emit("unit_moved", unit, x, y, from_x, from_y)
        if has_win:
            self.emit("player_wins", has_win)

