#!/usr/bin/python
# -*- coding:Utf-8 -*-

import game
import globals
import units
import unittest

class GameTest(unittest.TestCase):

    def test_game_init(self):
        "Game init correctly"
        game.Game()

    def test_position_is_valid_for_white_base(self):
        "check if position_is_valid work correctly for the white base"
        self.game = game.Game()

        for i in range (globals.WHITE_X, globals.WHITE_XW):
            for j in range(1, globals.BASE_H):
                result = globals.position_is_valid(i, j)
                self.assertEqual(result, True)

    def test_position_is_valid_for_black_base(self):
        "check if position_is_valid work correctly for the black base"
        self.game = game.Game()

        for i in range (globals.BLACK_X, globals.BLACK_XW):
            for j in range(1, globals.BASE_H):
                result = globals.position_is_valid(i, j)
                self.assertEqual(result, True)

    def test_position_is_valid_for_battlefield(self):
        "check if position_is_valid work correctly for the battlefield"
        self.game = game.Game()

        for i in range (globals.BATTLEFIELD_X, globals.BATTLEFIELD_XW):
            for j in range(1, globals.BASE_H):
                result = globals.position_is_valid(i, j)
                self.assertEqual(result, True)

    def test_position_is_valid_for_bads_value_for_Y_axe(self):
        "check if position_is_valid return false for bad value for Y axe"
        self.game = game.Game()

        bad = [i for i in range(-10, 0)]
        for i in [i for i in range(globals.BASE_H + 1, globals.BASE_H + 12)]:
            bad.append(i)

        for i in bad:
            self.assertEqual(globals.position_is_valid(1, i), False)

    def test_position_is_valid_for_bads_value_for_X_axe(self):
        "check if position_is_valid return false for bad value for X axe"
        self.game = game.Game()

        bad = [i for i in range(-10, globals.WHITE_X)]
        for i in [i for i in range(globals.BLACK_XW, globals.BLACK_XW + 10)]:
            bad.append(i)

        bad.append("globals.WHITE_XW")
        bad.append("globals.BLACK_XW")
        bad.append("globals.BATTLEFIELD_XW")

        for i in bad:
            self.assertEqual(globals.position_is_valid(i, 1), False)

    def test_board_white_base(self):
        "check if board function of game return correct values for white base"
        self.game = game.Game()

        for i in range(globals.WHITE_X, globals.WHITE_XW):
            for j in range(0, globals.BASE_H):
                self.assertEqual(self.game.get_unit(i, j), 0)

    def test_board_black_base(self):
        "check if board function of game return correct values for black base"
        self.game = game.Game()

        for i in range(globals.BLACK_X, globals.BLACK_XW):
            for j in range(0, globals.BASE_H):
                self.assertEqual(self.game.get_unit(i, j), 0)

    def test_board_battlefield(self):
        "check if board function of game return correct values for battlefield"
        self.game = game.Game()

        for i in range(globals.BATTLEFIELD_X, globals.BATTLEFIELD_XW):
            for j in range(0, globals.BASE_H):
                self.assertEqual(self.game.get_unit(i, j), 0)

    def test_end_turn(self):
        "test if end turn do things correctly"
        self.game = game.Game()
        self.game.new_game()

        turn_number = self.game.turn_number
        self.assertEqual("White", self.game.turn)

        self.game.end_turn()

        self.assertEqual([], self.game.has_moved)
        self.assertEqual([], self.game.chosen_type_of_units)
        self.assertEqual("Black", self.game.turn)
        self.assertEqual(turn_number, self.game.turn_number)

        self.game.end_turn()

        self.assertEqual(turn_number + globals.TURN, self.game.turn_number)

    def test_initial_placement_white_base(self):
        "test if the units are well placed for the white base"
        self.game = game.Game()
        self.game.new_game()

        units = self.game.initial_positions["White"]
        position = []

        for unit_class, positions in units.items():
            for x, y in positions:
                self.assertNotEqual(globals.EMPTY,
                        self.game.board.board[x][y])
                position.append((x, y))

        for i in range(globals.WHITE_X, globals.WHITE_XW):
            for j in range(0, globals.BASE_H):
                if (i, j) not in position:
                    self.assertEqual(globals.EMPTY,
                            self.game.board.board[i][j])

    def test_initial_placement_black_base(self):
        "test if the units are well placed for the black base"
        self.game = game.Game()
        self.game.new_game()

        units = self.game.initial_positions["Black"]
        position = []

        for unit_class, positions in units.items():
            for x, y in positions:
                self.assertNotEqual(globals.EMPTY,
                        self.game.board.board[x][y])
                position.append((x, y))

        for i in range(globals.BLACK_X, globals.BLACK_XW):
            for j in range(0, globals.BASE_H):
                if (x, y) not in position:
                    self.assertEqual(globals.EMPTY,
                            self.game.board.board[x][y])


    def test_board_add_unit_exception(self):
        "add unit should raise a exception when the case isn't empty"
        self.board = game.Board(5, 5)
        self.board.board[0][0] = 1 # not empty

        self.assertRaises(globals.CaseNotEmpty, self.board.add_unit,
                0, 0, "unit")

    def test_board_remove_unit_exception(self):
        "remove unit should raise a exception when the case is empty"
        self.board = game.Board(5, 5)

        for i in self.board.board:
            for j in i:
                self.assertRaises(globals.CaseIsEmpty, self.board.remove_unit,
                        0, 0)

    def test_game_return_to_base_exception(self):
        "return to base (for saboteur) should raise a exception when the base"\
        "is full"

        self.game = game.Game()
        self.saboteur_black = units.Saboteur("Black")
        self.saboteur_white = units.Saboteur("White")

        for i in range(1, 5):
            self.game.board.board[0][i] = 1 # not empty
            self.game.board.board[26][i] = 1

        for i in range(4):
            self.assertRaises(globals.BaseIsAnormalyFull,
                    self.game._return_to_base, self.saboteur_black)
            self.assertRaises(globals.BaseIsAnormalyFull,
                    self.game._return_to_base, self.saboteur_white)

if __name__ == "__main__":
    unittest.main()

# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
