from game import Board
from units import *
import globals

import gobject
import dbus, dbus.service
from dbus.mainloop.glib import DBusGMainLoop
import logging
logger = logging.getLogger('halebarde.server')

DBusGMainLoop(set_as_default=True)

server_interface = 'org.pygame.Pyhalebarde.server'

class RemoteGameServer(dbus.service.Object):
    """Export the functions and signals of game on dbus"""

    def __init__(self, bus, object_path, game):
        dbus.service.Object.__init__(self, bus, object_path)
        self.game = game
        self.unit_id = {}

        self.game.connect("initial_placement", self.initial_placement_cb)
        self.game.connect("new_game", self.new_game_cb)
        self.game.connect("unit_moved", self.unit_moved_cb)
        self.game.connect("saboteur_return", self.saboteur_return_cb)
        self.game.connect("ranged_attack", self.ranged_attack_cb)
        self.game.connect("direct_attack", self.direct_attack_cb)
        self.game.connect("player_wins", self.player_wins_cb)
        self.game.connect("turn_ended", self.turn_ended_cb)

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='iiii',out_signature='')
    def attack(self, to_x, to_y, from_x, from_y):
        self.game.attack(to_x, to_y, from_x, from_y)

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='iiii',out_signature='')
    def move(self, to_x, to_y, from_x, from_y):
        self.game.move(to_x, to_y, from_x, from_y)

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='ii',out_signature='b')
    def can_act(self, x, y):
        return self.game.can_act(x, y)

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='ii',out_signature='a(ii)')
    def can_attack_from(self, x, y):
        return self.game.can_attack_from(x, y)

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='ii',out_signature='a(ii)')
    def can_go_from(self, x, y):
        return self.game.can_go_from(x, y)

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='',out_signature='')
    def end_turn(self):
        self.game.end_turn()

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='',out_signature='')
    def new_game(self):
        self.game.new_game()

    @dbus.service.method(dbus_interface=server_interface,
                            in_signature='ii',out_signature='ss')
    def get_unit(self, x, y):
        return self.game.get_unit(x, y)

#CB
    def turn_ended_cb(self, game, turn, turn_number):
        logger.debug("Turn ended")
        self.TurnEnded(turn, turn_number)

    @dbus.service.signal(server_interface, signature='si')
    def TurnEnded(self, turn, turn_number):
        pass

    def new_game_cb(self, game):
        logger.debug("Game started")
        self.NewGame()

    @dbus.service.signal(server_interface, signature='')
    def NewGame(self):
        pass

    def player_wins_cb(self, game, player):
        logger.debug("%s wins." % player)
        self.PlayerWins(player)

    @dbus.service.signal(server_interface)
    def PlayerWins(self, player):
        pass

    def initial_placement_cb(self, game, unit, gx, gy):
        logger.debug("Unit %s placed at (%i, %i)"\
                % (unit, gx, gy))
        #attribute a id to the unit 
        id = len(self.unit_id)
        self.unit_id[unit] = id
        unit_class = unit.__class__.__name__
        self.InitialPlacement(id, unit_class, unit.side , gx, gy)

    @dbus.service.signal(server_interface, signature='issii')
    def InitialPlacement(self, unit_id, unit_class, team, gx, gy):
        pass

    def unit_moved_cb(self, game, unit, gx, gy, from_gx, from_gy):
        logger.debug("Unit %s moved from (%i, %i) to (%i, %i)"\
                % (unit, from_gx, from_gy, gx, gy))
        id = self.unit_id[unit]
        self.UnitMoved(id, gx, gy, from_gx, from_gy)
        
    @dbus.service.signal(server_interface, signature='iiiii')
    def UnitMoved(self, id, gx, gy, from_gx, from_gy):
        pass

    def saboteur_return_cb(self, game, unit, to_x, to_y, from_x, from_y):
        logger.debug("%s returned to home base at (%i,%i) from  (%i, %i)"\
                % (unit, to_x, to_y, from_x, from_y))
        id = self.unit_id[unit]
        self.SaboteurReturn(id, to_x, to_y, from_x, from_y)
        
    @dbus.service.signal(server_interface, signature='iiiii')
    def SaboteurReturn(self, unit, to_x, to_y, from_x, from_y):
        pass

    def ranged_attack_cb(self, game, unit, to_x, to_y, victim, from_x, from_y):
        logger.debug("Ranged attack from (%i,%i) by %s on (%i, %i)"\
                % (from_x, from_y, unit, to_x, to_y))
        id = self.unit_id[unit]
        victim_id = self.unit_id[victim]
        self.RangedAttack(id, to_x, to_y, victim_id, from_x, from_y)
        
    @dbus.service.signal(server_interface, signature='iiiiii')
    def RangedAttack(self, id, to_x, to_y, victim_id, from_x, from_y):
        pass
        
    def direct_attack_cb(self, game, unit, to_x, to_y, victim, from_x, from_y):
        logger.debug("Direct attack from (%i,%i) by %s on (%s, %s)"\
                % (from_x, from_y, unit, to_x, to_y))
        id = self.unit_id[unit]
        victim_id = self.unit_id[victim]
        self.DirectAttack(id, to_x, to_y, victim_id, from_x, from_y)
        
    @dbus.service.signal(server_interface, signature='iiiiii')
    def DirectAttack(self, id, to_x, to_y, victim_id, from_x, from_y):
        pass

class RemoteGameClient(gobject.GObject):
    """This class provide the same functions as the Game class and map the dbus
    signals to the gobject signals, so that the UI can use it as if it were the
    real Game class"""

    # Create some signal, which are fired when the state of the game change
    # these MUST BE the same as the one from the Game class !
    __gsignals__ = {
        # team which won
        "player_wins" : (gobject.SIGNAL_RUN_FIRST, None, [str]),
        #  New game started
        "new_game" : (gobject.SIGNAL_RUN_FIRST, None, []),
        # playing team, turn number
        "turn_ended" : (gobject.SIGNAL_RUN_FIRST, None, [str, int]),
        # unit, team, x, y
        "initial_placement": (gobject.SIGNAL_RUN_FIRST, None, [object, int,
            int]),
        # unit, to_x, to_y, from_x, from_y
        "unit_moved" : (gobject.SIGNAL_RUN_LAST, None, [object, int, int, int,
            int]),
        # attacker, to_x, to_y, victim, from_x, from_y
        "direct_attack" : (gobject.SIGNAL_RUN_LAST, None, [object, int, int,
            object, int, int]),
        # attacker, to_x, to_y, victim, from_x, from_y
        "ranged_attack" : (gobject.SIGNAL_RUN_LAST, None, [object, int, int,
            object, int, int]),
        # unit, to_x, to_y, from_x, from_y
        "saboteur_return": (gobject.SIGNAL_RUN_LAST, None, [object, int, int,
            int, int])
        }

    def __init__(self, server):
        gobject.GObject.__init__(self) # Call the parent constructeur
        self.server = server
        self._create_board(globals.TOTAL_WIDTH, globals.TOTAL_HEIGHT)
        self.turn = ''
        self.id_unit = {}

        connect = lambda signal, handler: server.connect_to_signal(signal,\
                                    handler, dbus_interface=server_interface)

        connect("TurnEnded", self.TurnEnded)
        connect("NewGame", self.NewGame)
        connect("PlayerWins", self.PlayerWins)
        connect("UnitMoved", self.UnitMoved)
        connect("InitialPlacement", self.InitialPlacement)
        connect("SaboteurReturn", self.SaboteurReturn)
        connect("RangedAttack", self.RangedAttack)
        connect("DirectAttack", self.DirectAttack)

    def _create_board(self, width, height):
        # This board serve as a cache so we don't have to make a network
        # round trip to know what is under the mouse.

        self.board = [globals.EMPTY] * width
        for  w in range(0, width): # we fill all the colon of the board
            self.board[w] = [globals.EMPTY] * height

    def attack(self, to_x, to_y, from_x, from_y):
        self.server.attack(to_x, to_y, from_x, from_y, dbus_interface=server_interface)

    def move(self, to_x, to_y, from_x, from_y):
        self.server.move(to_x, to_y, from_x, from_y, dbus_interface=server_interface)

    def can_act(self, x, y):
        return self.server.can_act(x, y, dbus_interface=server_interface)

    def can_attack_from(self, x, y):
        return self.server.can_attack_from(x, y, dbus_interface=server_interface)

    def can_go_from(self, x, y):
        return self.server.can_go_from(x, y, dbus_interface=server_interface)

    def new_game(self):
        self.server.new_game(dbus_interface=server_interface)

    def end_turn(self):
        self.server.end_turn(dbus_interface=server_interface)

    def get_unit(self, x, y):
        unit = self.board[x][y]
        return unit

#CB
    def TurnEnded(self, turn, turn_number):
        self.turn = turn
        self.emit("turn_ended", turn, turn_number)

    def NewGame(self):
        self.emit("new_game")

    def PlayerWins(self, player):
        self.emit("player_wins", player)
        self._create_board()

    def InitialPlacement(self, id, unit_class, team, x, y):
        unit = eval("%s('%s')" %(unit_class, team)) 
        self.id_unit[id] = unit
        self.board[x][y] = unit
        self.emit("initial_placement", unit, x, y)

    def UnitMoved(self, id, x, y, from_x, from_y):
        unit = self.id_unit[id]
        self.board[x][y] = unit
        self.board[from_x][from_y] = None
        self.emit("unit_moved", unit, x, y, from_x, from_y)

    def SaboteurReturn(self, id, to_x, to_y, from_x, from_y):
        unit = self.id_unit[id]
        self.board[to_x][to_y] = unit
        self.board[from_x][from_y] = None
        self.emit("saboteur_return", unit, to_x, to_y, from_x, from_y)

    def RangedAttack(self, id, to_x, to_y, victim_id, from_x, from_y):
        unit = self.id_unit[id]
        victim = self.id_unit[victim_id]
        self.board[to_x][to_y] = None
        self.emit("ranged_attack", unit, to_x, to_y, victim, from_x, from_y)

    def DirectAttack(self, id, to_x, to_y, victim_id, from_x, from_y):
        unit = self.id_unit[id]
        victim = self.id_unit[victim_id]
        self.board[to_x][to_y] = unit
        self.board[from_x][from_y] = None
        self.emit("direct_attack", unit, to_x, to_y, victim, from_x, from_y)
