#!/usr/bin/python
"Telepathy support for pyhalebarde"
import dbus
import logging

from dbus.service import method, signal, Object
import gobject

from halebarde.network import RemoteGameServer, RemoteGameClient
from halebarde.connection_watcher import ConnectionWatcher

from telepathy.interfaces import CHANNEL_TYPE_CONTACT_LIST,\
    CONNECTION_INTERFACE_REQUESTS, CONNECTION_INTERFACE_SIMPLE_PRESENCE,\
    CHANNEL_INTERFACE_GROUP, CHANNEL_INTERFACE, CHANNEL_INTERFACE_TUBE,\
    CHANNEL_TYPE_DBUS_TUBE
from telepathy.constants import HANDLE_TYPE_LIST, HANDLE_TYPE_CONTACT,\
    CONNECTION_PRESENCE_TYPE_AVAILABLE, CONNECTION_PRESENCE_TYPE_AWAY,\
    CONNECTION_PRESENCE_TYPE_EXTENDED_AWAY, CONNECTION_PRESENCE_TYPE_BUSY,\
    TUBE_STATE_OPEN, SOCKET_ACCESS_CONTROL_CREDENTIALS, TUBE_CHANNEL_STATE_REMOTE_PENDING
from telepathy.client.channel import Channel
from telepathy.client.conn import Connection

import halebarde.ui

#FIXME : Should be passed as an arg instead of being a global variable
TUBE_SERVICE = "org.pygame.Pyhalebarde"

def search_contact_from_id(contact, conn):
    """Return the handle that match the id in given connexion"""

    if CONNECTION_INTERFACE_REQUESTS not in conn:
        logging.warning("CONNECTION_INTERFACE_REQUESTS not implemented in %s" %\
                (conn.service_name))
        return None

    if not conn_can_request_contact_list(conn):
        logging.warning("CHANNEL_TYPE_CONTACT_LIST not requestable in %s" %\
                (conn.service_name))
        return None

    # get contacts list
    yours, path, props = conn[CONNECTION_INTERFACE_REQUESTS].EnsureChannel({
        'org.freedesktop.Telepathy.Channel.ChannelType': CHANNEL_TYPE_CONTACT_LIST,
        'org.freedesktop.Telepathy.Channel.TargetHandleType': HANDLE_TYPE_LIST,
        'org.freedesktop.Telepathy.Channel.TargetID': 'subscribe'})

    chan = Channel(conn.service_name, path)
    contacts = chan[CHANNEL_INTERFACE_GROUP].GetMembers()
    ids = conn.InspectHandles(HANDLE_TYPE_CONTACT, contacts)

    for handle, id in zip(contacts, ids):
        if id == contact:
            return handle

def contact_is_online(handle, conn):
    presences = conn[CONNECTION_INTERFACE_SIMPLE_PRESENCE].GetPresences([handle])

    type, status, msg = presences[handle]
    return type in [CONNECTION_PRESENCE_TYPE_AVAILABLE,
            CONNECTION_PRESENCE_TYPE_AWAY,
            CONNECTION_PRESENCE_TYPE_EXTENDED_AWAY, CONNECTION_PRESENCE_TYPE_BUSY]

CHANNEL_TYPE = 'org.freedesktop.Telepathy.Channel.ChannelType'
TARGET_HANDLE_TYPE = 'org.freedesktop.Telepathy.Channel.TargetHandleType'

def conn_can_request_contact_list(conn):
    rcc = conn['org.freedesktop.DBus.Properties'].Get(CONNECTION_INTERFACE_REQUESTS,
            'RequestableChannelClasses')
    for channel_class in rcc:
        fixed_properties, allowed_properties = channel_class
        if fixed_properties[CHANNEL_TYPE] ==  CHANNEL_TYPE_CONTACT_LIST\
            and fixed_properties[TARGET_HANDLE_TYPE] == HANDLE_TYPE_LIST :
            # We found the correct channel type and handle
            return True
    return False

def offer_tube(conn, handle):
    """Offer a dtube to contact designed by @handle"""
    # FIXME: check if contact supports our tube
    yours, path, props = conn[CONNECTION_INTERFACE_REQUESTS].EnsureChannel({
        'org.freedesktop.Telepathy.Channel.ChannelType': CHANNEL_TYPE_DBUS_TUBE,
        'org.freedesktop.Telepathy.Channel.TargetHandleType': HANDLE_TYPE_CONTACT,
        'org.freedesktop.Telepathy.Channel.TargetHandle': handle,
        CHANNEL_TYPE_DBUS_TUBE + '.ServiceName': TUBE_SERVICE})

    def tube_state_changed_cb(state):
        if state == TUBE_STATE_OPEN:
            print "Tube accepted "

            tube_conn = dbus.connection.Connection(address)
            # Create UI and server
            
            ui = halebarde.ui.get_ui()
            w = ui.new_game_window(side = "White", tube = chan)
            game = w.board.game
            p = RemoteGameServer(tube_conn, "/Game", game)
            print "Serveur created"
        elif state == TUBE_CHANNEL_STATE_REMOTE_PENDING:
            print "Other player is invited, waiting for response..."
        else:
            print "Tube state changed", state

    def tube_closed_cb():
        print "tube closed"

    chan = Channel(conn.service_name, path)
    chan[CHANNEL_INTERFACE_TUBE].connect_to_signal('TubeChannelStateChanged',
            tube_state_changed_cb)
    chan[CHANNEL_INTERFACE].connect_to_signal('Closed', tube_closed_cb)
    print "Invited contact on %s" %TUBE_SERVICE
    address = chan[CHANNEL_TYPE_DBUS_TUBE].Offer({}, SOCKET_ACCESS_CONTROL_CREDENTIALS)

def offer_tube_to_contact_cb(watcher, conn, contact):
    #TODO : Remove itself when contact is found.
    handle = search_contact_from_id(contact, conn)
    if handle: 
        print "Contact %s found in %s" % (contact, conn.service_name)
        if contact_is_online(handle, conn):
            print "contact online"
            offer_tube(conn, handle)
    else:
        print "Contact %s not found in %s" % (contact, conn.service_name)

class Handler(Object):
    @dbus.service.method("org.gnome.Empathy.TubeHandler",
            in_signature='soouu', out_signature='')
    def HandleTube(self, bus_name, connection, channel, handle_type, handle):
        #We need a try here cause strangly it doesn't display exceptions
        try:
            logging.info("Handling a new tube!")
            # TODO: ask to the user if he wants to accept or decline the tube

            conn = Connection(bus_name, connection)
            chan = Channel(bus_name, channel)
            self.chan = chan

            self.address = chan[CHANNEL_TYPE_DBUS_TUBE].Accept(SOCKET_ACCESS_CONTROL_CREDENTIALS)
            chan[CHANNEL_INTERFACE_TUBE].connect_to_signal(\
                'TubeChannelStateChanged', self.tube_state_changed_cb)
        except Exception, e:
            print e
            raise

    def tube_state_changed_cb(self, state):
        if state == TUBE_STATE_OPEN:
            print "Tube opened"
            tube_conn = dbus.connection.Connection(self.address)

            game_server = tube_conn.get_object(object_path="/Game")
            game = RemoteGameClient(game_server)
            print game
            ui = halebarde.ui.get_ui()
            ui.new_game_window(game, "Black", tube = self.chan)
        else:
            print "Tube state changed", state
 
def create_handler(tube_service):
    """create a D-Bus handler service,
    which wait for someone to offer us a tube"""
    logging.info("Creating the handler service")
    # Open the local session bus
    bus = dbus.SessionBus()
    handler_name = "org.gnome.Empathy.DTubeHandler.%s" % tube_service.replace('.', '_')
    handler_path = "/org/gnome/Empathy/DTubeHandler/%s" % tube_service.replace('.','_')
    logging.info("Bus name : %s" % handler_name)
    logging.info("Bus path : %s" % handler_path)
    name = dbus.service.BusName(handler_name, bus)
    handler = Handler(bus, handler_path)
    return name, handler

def invite_contact(contact):
    watcher = ConnectionWatcher()
    print "Looking for contact..."
    watcher.connect('connection-added', offer_tube_to_contact_cb, contact)
